#!/usr/bin/env python
# coding=utf-8

from dsmapi2 import *
from client import initLib, cleanupLib
from classes import *
from ds_exceptions import *
from storage import intern_DSID

import logging

TIMESTMP_FMT = ("%H:%M:%S") # no %f for msecs
FORMAT = '%(asctime)s.%(msecs)03d %(name)-5s %(message)s'
logging.basicConfig(format=FORMAT, datefmt=TIMESTMP_FMT, level=logging.INFO)

log = logging.getLogger("StructureAPI")

def check_str(test_str):
    import re
    #http://docs.python.org/library/re.html
    #re.search returns None if no position in the string matches the pattern
    #pattern to search for any character other then . a-z 0-9
    pattern = r'[^\.a-z0-9]'
    return not re.search(pattern, test_str)

def fetch_all_zones(h, dsmid):
    zoneInfo = fetch_ZoneInfos(h, dsmid)
    return map(build_Zone, zoneInfo)

def build_Zone(zoneInfo):

    # unused:
    # dsmLocalId
    # numGroups

    zone = Zone(zoneInfo.zoneId, zoneInfo.name)
    return zone

class ZoneInfo:
    def __init__(self, zoneInfoRecord):
        (self.zoneId, self.dsmLocalId, self.numGroups, self.name) = zoneInfoRecord

    def __str__(self):
        return "zone : %5d name: %s" % (self.zoneId, self.name)
    def __repr__(self):
        return "zone : %(zoneid)5d name: %(name)s localId: %(dsmLocalId)d groups: %(numGroups)d" \
                % self.__dict__

def fetch_ZoneInfos(h, dsmid):
    assert(type(dsmid) == list)
    ret = ZoneCount(h, dsmid)
    if extractError(ret) != 0:
        raise createDsException(extractError(rec), "ZoneCount")

    ct = ret[1]
    zoneInfo = []
    for i in range(ct):
        ret = ZoneInfo_by_index(h, dsmid, i)
        if extractError(ret) != 0:
            raise createDsException(extractError(rec), "ZoneInfo_by_index")

        zoneInfo.append(ZoneInfo(ret[1:]))

    return zoneInfo

def build_Group(groupInfo):
    pass

class GroupInfo:
    def __init__(self, rec, zoneId, dsmid):
        (self.groupId, self.stateMachineGroupId, self.numDevices, self.name,
         self.stopTimeout, self.remoteStopTimeout) = rec
        self.zoneId = zoneId
        self.dsmid = dsmid
        #if (check_str(self.name)):
        #    self.name = "<unknown>"

    def __repr__(self):
        return \
"group: %(groupId)3d name: %(name)20s zone: %(zoneId)5d dsm: %(dsmid)s \
    devices: %(numDevices)d SM: %(stateMachineGroupId)d timeout %(stopTimeout)d remote-timeout %(remoteStopTimeout)d" \
            % (self.__dict__)

def fetch_GroupInfos(h, dsmid, zoneId, numGroups):

    infos = []
    for i in range(numGroups):
        ret = ZoneGroupInfo_by_index(h, dsmid, zoneId, i)
        if extractError(ret) != 0:
            raise createDsException(extractError(ret), "ZoneGroupInfo_by_index")

        infos.append(GroupInfo(ret[1:], zoneId, dsmid))

    return infos

def fetch_GroupInfo(h, dsmid, zoneId, groupId):
    ret = ZoneGroupInfo_by_id(h, dsmid, zoneId, groupId)
    if extractError(ret) != 0:
        raise createDsException(extractError(ret), "ZoneGroupInfo_by_id")

    return GroupInfo(ret[1:], zoneId, dsmid)

#afenkart: mit dem ltmode hab ich probleme, der gehoert in den output,  ist es
# gegeben dass es immer einen output gibt?
#chitz: DSDevice: wenn einen EAN gespeichert ist (Bank 3, 0x2A ff) dann die
# info sonst aus dem terminal block
#chitz: bis heute gibt es jeweils nur einen output
#chitz: wie wir das in zukunft genau machen haben wir noch nicht definiert...
# chitz: name?: jein, nur wenn keine EAN definiert ist und dsdevice aus
#   genau einem terminal-block besteht
# afenkart: ansonsten -> trash
def create_base_device(deviceInfoRecord, dsmid, h):
    (shortAddress, vendorId, productId, functionId, revisionId, zoneId,
     active, locked, outputMode, ltMode, groups, name, serialNumber) = deviceInfoRecord

    firmwareRevision = revisionId # mtr: ja, das ist wohl dasselbe
    groupId = functionId >> 12
    connectionState = ConnectionState.active if active else ConnectionState.inactive

    # TODO vendorId, productId, functionId >> 12
    # TODO functionId & 0xfff
    # TODO locked
    # TODO ltMode
    # TODO name

    devSpec = dict(groups = groups)



    (err, name, hwinfo, gtin) = \
            DsmApiGetDeviceDescription(vendorId, productId, groupId,
                                       firmwareRevision)
    if err != 0:
        # TODO device description might not be found if firmwareRevision < 0x350
        raise createDsException(err, "DsmApiGetDeviceDescription(%s/%s/%s/%s/%s" %
                                    (vendorId, productId, groupId,
                                     firmwareRevision))

    (err, dsid) = DsmApiGetDeviceDSID(vendorId, productId, groupId, firmwareRevision,
                                      serialNumber)
    if err != 0:
        raise createDsException(err, "DsmApiGetDeviceDSID(%s/%s/%s/%s/%s" %
                                    (vendorId, productId, groupId,
                                     firmwareRevision, serialNumber))


    # missing: icon, room, sensor
    tb = TerminalBlock(shortAddress, firmwareRevision, zoneId, connectionState, \
                       serialNumber, dsmid, gtin, hwinfo)

    # missing: group, name, type_, mode conversion
    output = Output(outputMode)
    tb.setOutput(output)

    # nur of buttons, etc
    #int subclass = (m_FunctionID & 0xfc0) >> 6;
    #int funcmodule = (m_FunctionID & 0x3f);

    pb = create_input_push_button(h, dsmid, tb.shortAddress)
    tb.appendInputPushButton(pb)

    bs = create_input_binary_states(h, dsmid, tb.shortAddress) # list
    tb.appendInputPushButton(bs)

    devSpec['terminalBlocks'] = [tb]

    if not active:
        # TODO, can't fetch EAN -> hence no gtin for DSDevice, mark as pending
        devSpec['dsid'] = dsid
        devSpec['serialNumber'] = serialNumber
        devSpec['oemAreaState'] = OemState.fetching
        return (DSDevice(**devSpec), tb)

    # TODO dsdevice.EEPROM


    # TODO if adding device without GTIN/EAN we must later remove it, if the TB
    # belongs to an already existing multi-TB device

    # TODO product handler callback must fetch EAN area once TB becomes active

    # TODO groupid == DSDevice.isStandardGroup?

    # catch Exception, device disconnected

    # missing: icon, productInfo, name, dsid
    dev = DSDevice(**devSpec)
    return (dev, tb)


SETLOCALPRIORITY    = 0x1
NOCALLSPRESENT      = 0x2 # aka no implicit coming home call
WHOLE_APARTMENT     = 0x4
def create_input_push_button(h, dsmid, shortAddress):
    ret = DeviceButtonInfo_by_device(h, dsmid, shortAddress)
    (err, buttonId, groupMembership, activeGroup, flags) = ret
    if (err != 0):
        raise Exception("DeviceButtonInfo_by_device %d" % err)

    # TODO buttonId, groupMembership

    setsLocalPriority = bool(flags & SETLOCALPRIORITY)
    callsPresent = not bool(flags & NOCALLSPRESENT)
    #wholeApartment = bool(flags & WHOLE_APARTMENT)

    #print "activeGroup", activeGroup

    # missing: mode, name, ltMode
    return InputPushButton(activeGroup, setsLocalPriority, callsPresent)


# TODO unit test for AKM device
def create_input_binary_states(h, dsmid, shortAddress):
    (err, ct) = DeviceBinaryInput_get_count(h, dsmid, shortAddress)
    if (err == ERROR_WRONG_MSGID or err == ERROR_WRONG_MODIFIER):
        # dSM FW does not yet know about BinaryInputs, aka AKM inputs
        # -> upgrade dSM
        # TODO catch and swallow this error, don't crash the app
        log.warning("DeviceBinaryInput_get_count: mismatch between DSM firmware and API library")
        return
    elif (err != 0):
        raise createDsException(err, "DeviceBinaryInput_get_count")

    devices = list()
    for i in range(ct):
        (err, targetGroupType, targetGroup, inputType, inputId) = \
                DeviceBinaryInput_get_by_index(h, dsmid, shortAddress, i)

        if (err != 0):
            raise createDsException("DeviceBinaryInput_get_by_index")

        # missing: mode, ltMode, name
        binaryInput = BinaryInput(targetType, targetGroup, inputType, inputId)
        devices.append(binaryInput)

    return devices

"- updateButtonGroupFromMeter(dsmDSID, result) \
- DeviceButtonInfo_by_device(m_DSMApiHandle, _dsMeterID, _spec.ShortAddress,... \
- updateBinaryInputTableFromMeter(dsmDSID, result); \
- DeviceBinaryInput_get_count(m_DSMApiHandle, _dsMeterID, _spec.ShortAddress,... \
- loop: DeviceBinaryInput_get_by_index(m_DSMApiHandle, _dsMeterID, _spec.ShortAddress,.."
def fetch_all_devices(h, dsmid):
    devices = []
    (error, ct) = ZoneDeviceCount_all(h, dsmid, 0)
    if error != 0:
        raise Exception("ZoneDeviceCount_all %d" % ret[0])
    for i in range(ct):
        try:
            ret = DeviceInfo_by_index(h, dsmid, 0, i)
            if ret[0] != 0:
                raise Exception("DeviceInfo_by_index %d" % ret[0])
            (dev, tb) = create_base_device(ret[1:], dsmid, h)
            devices.append(dev, tb)
        except Exception, e:
            print "ignoring device ", ret[1:]

    return devices


def fetch_device(h, dsmid, shortAddress):
    ret = DeviceInfo_by_device_id(h, dsmid, shortAddress)
    if ret[0] != 0:
        raise Exception("DeviceInfo_by_device_id %d" % ret[0])

    (dev, tb) = create_base_device(ret[1:], dsmid, h)
    return (dev, tb)

def fetch_all_dsmid(h):
    # TODO if isolated returns, 0
    rec = DsmApiGetBusMembers(h, 60)
    err = extractError(rec)
    if err < 0: # err >= 0 => number of bus members
        raise createDsException(extractError(rec), "DsmApiGetBusMembers")
    if err == 0:
        log.info("No DSM found\n");
        return []
    l = filter(lambda x: DsmApiIsdSM(x), rec[1:])
    return map(intern_DSID, l)

def build_DSM(h, dsid):
    rec = dSMInfo(h, dsid)
    if extractError(rec) != 0:
        raise createDsException(extractError(rec), "dSMInfo")

    # dSMProperties_get_flags
    return DSMeter(rec[1:])

def sync_DSM(h, dsm):
    # TODO implement
    return build_DSM(h, dsm.dsid)

def getDsmConfigHash(h, dsid):
    rec = dSMConfig_get_hash(h, dsid)
    if extractError(rec) != 0:
        raise createDsException(extractError(rec), "isDsmConfigChanged")

    (dataModelHash, configCount, binaryInputEventCount) = rec[1:]
    return (dataModelHash, configCount, binaryInputEventCount)


if __name__== "__main__":
    global h

    h = initLib("tcp://localhost:8442")

    print "DSM:"
    dSMs = []
    for dsmid in fetch_all_dsmid(h):
        dsm = build_DSM(h, dsmid)
        print "%r\n" % build_DSM(h, dsmid)
        dSMs.append(dsm)

    print "zones"
    dsm1 = dSMs[0]
    zoneInfos = fetch_ZoneInfos(h, dsm1.dsid)
    for zInfo in zoneInfos:
        build_Zone(zInfo)
        print zInfo
    print

    zone1 = zoneInfos[0]
    print "groups in %s " % str(zone1)
    groupInfos = fetch_GroupInfos(h, dsm1.dsid, zone1.zoneId, zone1.numGroups)
    for gInfo in groupInfos:
        #build_Group(info)
        print gInfo
    print

    print "devices"
    devices = fetch_all_devices(h, dsm1.dsid)
    for x in devices:
        print x
    print

    getDsmConfigHash(h, dsm1.dsid)

    cleanupLib(h)
