#!/usr/bin/env python
from ds485_const import *
from notification_queue import *
from process_pkt import split_pkt
from util import list2hexstr

import sys, os
import asyncore, socket
import traceback, logging
import time
import errno

TIMESTMP_FMT = ("%H:%M:%S") # no %f for msecs
FORMAT = '%(asctime)s.%(msecs)03d %(name)-5s %(message)s'
logging.basicConfig(format=FORMAT, datefmt=TIMESTMP_FMT, level=logging.INFO)

class ClientState:
    connected = False
    dsmids = []

    def dump_device_list(self):
        log = logging.getLogger("Client-%d" % self.ct)
        log.info("client device list")
        for dsm in self.dsmids:
            log.info("\t%s", dsm)

class Client(asyncore.dispatcher):
    ct = 0

    def __init__(self, host, port, state):
        try:
            asyncore.dispatcher.__init__(self)
            log = logging.getLogger("Client-%d" % self.ct)
            self.ct += 1
            self.log = log
            log.info("connect %s:%d", host, port)
            self.queue = Queue()
            self.outbuf = []
            self.buf = []
            self.state = state
            state.ct = self.ct
            self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
            self.connect((host, port))
        except Exception, e:
            print traceback.format_exc()
            self.close()

    def handle_connect(self):
        try:
            log = self.log
            log.info("handle_connect")
            notification_queue.put((handle_connect, None, self))
        except Exception, e:
            print traceback.format_exc()

    def handle_exp(self):
        try:
            log = self.log
            log.info("handle_expt")
        except Exception, e:
            print traceback.format_exc()

    def handle_close(self):
        log = self.log
        log.info("handle_close")
        notification_queue.put((handle_close, None, self))
        self.close()

    def handle_read(self):
        try:
            log = self.log
            log.debug("handle_read")
            data = self.recv(8192)
            if not data:
                return
            buf = self.buf
            buf.extend(map(ord, list(data)))
            for pkt in split_pkt(buf):
                notification_queue.put((handle_read, pkt, self))
        except socket.error,e:
            if e[0] in [errno.EWOULDBLOCK, errno.EAGAIN]:
                # first read always fails with errno 11, mute it
                log.info("handle_read: EWOULDBLOCK EAGAIN")
                return
            elif e[0] in [errno.ECONNREFUSED]:
                log.info("handle_read: Server down or not yet ready")
                return
            print traceback.format_exc()
        except Exception, e:
            print e, " ", type(e), " e[0] ", e[0]
            print traceback.format_exc()

    def writable(self):
        log = self.log
        ret = (len(self.outbuf) > 0 or (not self.queue.empty()))
        if ret:
            log.debug("writable %r", ret)
        return ret

    def handle_write(self):
        log = self.log
        if ((len(self.outbuf) == 0) and self.queue.empty()):
            return

        if not self.outbuf:
            self.outbuf = self.queue.get()
        sent = self.send(self.outbuf)
        log.debug("handle_write sent:%r pkt:%r", sent, len(self.outbuf))
        self.outbuf = self.outbuf[sent:]

    def push(self, pkt):
        self.queue.put(pkt)
